/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_NODE_H__
#define __AIDGE_NODE_H__

#include <cassert>
#include <memory>
#include <set>
#include <string>
#include <vector>
#include <utility>

#include "graph/Connector.hpp"
#include "operator/Operator.hpp"
#include "utils/Types.h"

namespace Aidge {

using NodePtr = std::shared_ptr<Node>;

class GraphView;

class Node : public std::enable_shared_from_this<Node> {
private:
    /// @brief Name of the Node. Should be unique
    std::string mName; 

    /// @brief Set of pointers to GraphView (instances including this Node)
    std::set<std::shared_ptr<GraphView>> mViews = std::set<std::shared_ptr<GraphView>>(); 

    /// @brief Pointer to the associated Operator
    const std::shared_ptr<Operator> mOperator;

    /// @brief List of parent nodes (Parent --> Node --> Child)
    std::vector<NodePtr> mParents;

    /// @brief List of child nodes for each output (Parent --> Node --> Child)
    std::vector<std::vector<NodePtr>> mChildren;

    /// @brief InID of Child node.
    std::vector<std::vector<IOIndex_t>> mIdInChildren;

    /// @brief OutID of Parent node. Default: gk_IODefaultIndex.
    std::vector<IOIndex_t> mIdOutParents;

public:
    Node() = delete;
    Node(std::shared_ptr<Operator> op, const char *name = nullptr);

    virtual ~Node() = default;

    friend bool operator==(const Node &lhs, const Node &rhs) {
        return lhs.shared_from_this() == rhs.shared_from_this();
    }

    ///////////////////////////////////////////////////////
    //        FUNCTIONAL DESCRIPTION
    ///////////////////////////////////////////////////////

    Connector operator()(const std::vector<Connector> ctors);

    ///////////////////////////////////////////////////////
    //        INNER
    ///////////////////////////////////////////////////////

    /**
     * @brief Name of the node.
     * @return std::string
     */
    inline std::string name() const noexcept { return mName; }

    /**
     * @brief Set the node name.
     * @warning Undefined behaviour when several Nodes have the same name.
     * @param name New name for the node.
     */
    void setName(const std::string &name);

    /**
     * @brief Type of the node.
     * @return std::string
     */
    inline std::string type() const { return mOperator->type(); }

    ///////////////////////////////////////////////////////
    //        OPERATORS
    ///////////////////////////////////////////////////////

    /**
     * @brief Run forward() function of the associated Operator
     */
    void forward();

    /**
     * @brief Run backward() function of the associated Operator
     */
    void backward();

    /**
     * @brief Get the Operator object of the Node
     * @return std::shared_ptr<Operator>
     */
    inline std::shared_ptr<Operator> getOperator() const { return mOperator; }

    ///////////////////////////////////////////////////////
    //        TENSOR MANAGEMENT
    ///////////////////////////////////////////////////////

    /**
     * @brief Whether or not every input of the Node is linked to a Tensor.
     * If true then the Node is ready to be executed.
     * @return true
     * @return false
     */
    bool valid() const;

    /**
     * @brief List of pair <Parent, ID of the data intput>. When an input is not
     * linked to any Parent, the pair is <nullptr, gk_IODefaultIndex>.
     * @return std::vector<std::pair<NodePtr, IOIndex_t>>
     */
    std::vector<std::pair<NodePtr, IOIndex_t>> dataInputs() const;

    /**
     * @brief List of pair <Parent, ID of the intput>. When an input is not linked
     * to any Parent, the pair is <nullptr, gk_IODefaultIndex>.
     * @return std::vector<std::pair<NodePtr, IOIndex_t>>
     */
    std::vector<std::pair<NodePtr, IOIndex_t>> inputs() const;

    /**
     * @brief Parent and its output Tensor ID linked to the inID-th input Tensor.
     * If the input is not linked to any Parent, the pair is <nullptr, gk_IODefaultIndex>.
     * @param inID
     * @return std::pair<NodePtr, IOIndex_t>
     */
    inline std::pair<NodePtr, IOIndex_t> input(IOIndex_t inID) const {
        assert((inID != gk_IODefaultIndex) && (static_cast<IONb_t>(inID) < nbInputs()) && "Input index out of bound.");
        return std::pair<NodePtr, IOIndex_t>(mParents[inID], mIdOutParents[inID]);
    }

    /**
     * @brief Get the lowest index in the input Data Parent list equal to the
     * nullptr.
     * @return std::size_t
     */
    inline IOIndex_t getFirstFreeDataInput() const {
        IOIndex_t i = 0;
        for (; (static_cast<IONb_t>(i) < nbDataInputs()) && (input(i).second >= 0); ++i) {}
        // assert((i<nbDataInputs()) && "No free data input for Node");
        return (static_cast<IONb_t>(i) < nbDataInputs()) ? i : gk_IODefaultIndex;
    }

    IONb_t getNbFreeDataInputs() const;

    /**
     * @brief List input ids of children liked to outputs of the node
     * @return std::vector<std::vector<std::pair<NodePtr,
     * IOIndex_t>>>
     */
    std::vector<std::vector<std::pair<NodePtr, IOIndex_t>>> outputs() const;

    /**
     * @brief Children and their input Tensor ID linked to the outID-th output
     * Tensor.
     * @param outID
     * @return std::vector<std::pair<NodePtr, IOIndex_t>>
     */
    std::vector<std::pair<NodePtr, IOIndex_t>> output(IOIndex_t outID) const;

    /**
     * @brief Number of input specifically for data
     * @details [data, data, weight, bias] => 4
     * @return IOIndex_t
     */
    inline IONb_t nbInputs() const noexcept { return getOperator()->nbInputs(); }

    /**
     * @brief Number of input specifically for data
     * @details [data, data, weight, bias] => 2
     * @return IOIndex_t
     */
    inline IONb_t nbDataInputs() const noexcept {
        return getOperator()->nbDataInputs();
    }

    /**
     * @brief Number of inputs linked to a Parent's output.
     * @return IOIndex_t
     */
    IONb_t nbValidInputs() const;

    /**
     * @brief Getter for the number of Output Tensors of the Node.
     * @return IOIndex_t
     */
    inline IONb_t nbOutputs() const noexcept { return getOperator()->nbOutputs(); }

    IONb_t nbValidOutputs() const;

    ///////////////////////////////////////////////////////
    //        TOPOLOGY
    ///////////////////////////////////////////////////////

    /**
     * @brief Vector of pointers to each GraphView containing the object
     * @return std::vector<GraphView>
     */
    inline std::set<std::shared_ptr<GraphView>> views() const noexcept {
        return mViews;
    }

    /**
     * @brief Add a GraphView pointer to the list of GraphView containing
     * the current Node. This feature allows transparent GraphViews.
     * @param graphPtr Pointer to GraphView to add to the list.
     */
    inline void addView(const std::shared_ptr<GraphView> graphPtr) {
        mViews.insert(graphPtr);
    }

    inline void removeView(const std::shared_ptr<GraphView> graphPtr) {
        if (mViews.find(graphPtr) != mViews.end()) {
        mViews.erase(graphPtr);
        }
    }

    /**
     * @brief Link another Node to an output of the current Node.
     * @param otherNode Pointer to the other Node.
     * @param outId ID of the output Tensor to connect to the other Node.
     * Default to 0.
     * @param otherInId ID of the input Tensor to connect to the current Node.
     * Default to the first avaible data input.
     */
    void addChild(NodePtr otherNode,
                    const IOIndex_t outId = IOIndex_t(0),
                    IOIndex_t otherInId = gk_IODefaultIndex);

    /**
     * @brief Link a Node from a specific GraphView to the current Node.
     * @param otherView Pointer to the GraphView whose content should be
     * linked to the current Node.
     * @param outId ID of the output Tensor to connect to the other Node.
     * Default to 0.
     * @param otherInId Pair of pointer to Node and Tensor ID for specifying the
     * connection. If the GraphView whose content is linked has only one input
     * Node, then it defaults to the first available data input Tensor of this
     * Node.
     */
    void addChild(std::shared_ptr<GraphView> otherView,
                    const IOIndex_t outId = IOIndex_t(0),
                    std::pair<NodePtr, IOIndex_t> otherInId =
                    std::pair<NodePtr, IOIndex_t>(nullptr, gk_IODefaultIndex));

    /**
     * @brief Get the list of parent Nodes. As an input is linked to a unic Node,
     * if non is linked then the parent is a nullptr.
     * @return std::vector<NodePtr>
     */
    std::vector<NodePtr> getParents() const;

    inline NodePtr &getParents(IOIndex_t inID) {
        assert(inID != gk_IODefaultIndex);
        return mParents.at(inID);
    }

    NodePtr popParent(const IOIndex_t inID);

    void removeParent(IOIndex_t inID);

    /**
     * @brief Get the Children object. Children do not include any nullptr as
     * an output maybe linked to nobody and the Node would still work fine.
     * @return std::set<NodePtr>>
     */
    std::set<NodePtr> getChildren() const;

    std::vector<std::vector<NodePtr>> getOrderedChildren() const;

    std::vector<NodePtr> getChildren(IOIndex_t outID) const;

    void removeChild(IOIndex_t outID, std::size_t childId);

    /**
     * @brief Remove every link of surrounding nodes to it and conversly
     */
    void resetConnections(bool includeLearnableParam = false);

    private:
    ///////////////////////////////////////////////////////
    //        OPERATORS
    ///////////////////////////////////////////////////////

    // void setOperator(const std::shared_ptr<Operator> op_ptr);

    ///////////////////////////////////////////////////////
    //        TENSOR MANAGEMENT
    ///////////////////////////////////////////////////////

    void setInput(IOIndex_t inID, IOIndex_t newNodeOutID);

    ///////////////////////////////////////////////////////
    //        TOPOLOGY
    ///////////////////////////////////////////////////////

    /**
     * @brief add function specialized in adding Nodes.
     * @param other_node
     * @param outID
     * @param other_inID
     */
    void addChildOp(NodePtr other_node, const IOIndex_t outID,
                    IOIndex_t other_inID);

    /**
     * @brief add functon specialized in adding GraphView.
     *
     * @param other_graph
     * @param outID
     * @param other_inID
     */
    void addChildView(std::shared_ptr<GraphView> other_graph,
                        const IOIndex_t outID,
                        std::pair<NodePtr, IOIndex_t> other_inID);

    /**
     * @brief Add a Node to the list of parents.
     * @param other_node Node to add to parents list.
     * @param inID index for adding the parent.
     */
    void addParent(const NodePtr other_node, const IOIndex_t inID);
};
} // namespace Aidge

#endif /* __AIDGE_NODE_H__ */