/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_ADD_H__
#define __AIDGE_ADD_H__

#include <numeric>
#include <vector>
#include <cmath>

#include "utils/Parameter.hpp"
#include "utils/Registrar.hpp"
#include "operator/Operator.hpp"
#include "operator/Producer.hpp"
#include "data/Tensor.hpp"
#include "graph/Node.hpp"

namespace Aidge {

template <std::size_t NUM>
class Add_Op : public Operator,
    public Registrable<std::string, std::unique_ptr<OperatorImpl>(const Add_Op<NUM>&)> {
public:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, NUM> mInputs;
    std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char* Type = "Add";

    constexpr Add_Op()
            : Operator(Type),
            mOutput(std::make_shared<Tensor>())
    {
        assert(NUM > 0 && "Add should have at least one input");
        for (std::size_t i = 0; i<NUM; ++i) {
            mInputs[i] = std::make_shared<Tensor>();
        }
        setDatatype(DataType::Float32);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    constexpr void setInput(IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(static_cast<std::size_t>(inputIdx) < NUM && "wrong inputIdx for Add operator.");
        assert(strcmp(data->type(), Tensor::Type)==0 && "input data must be of Tensor type");

        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    constexpr void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            const auto expectedDims =  mInputs[0]->dims();
            std::size_t nonEmptyInputTensor = 1;
            for (; nonEmptyInputTensor<NUM && (!mInputs[nonEmptyInputTensor]->empty()); ++nonEmptyInputTensor) {
                assert(expectedDims == mInputs[nonEmptyInputTensor]->dims());
            }
            if (nonEmptyInputTensor == NUM) {
                mOutput->resize(expectedDims);
            }
        }
    }

    bool outputDimsForwarded() const override final {
        std::size_t forwarded = 0;
        for (; forwarded < NUM && (!mInputs[forwarded]->empty()); ++forwarded) {}
        return ((forwarded==NUM) && !(mOutput->empty()));
    }

    // void checkDims() const override final {
    //     assert(outputDimsForwarded());
    //     for (const auto& in : mInputs) {
    //         assert(in->dims() == mOutput->dims());
    //     }
    // }

    std::shared_ptr<Data> getInput(IOIndex_t inputIdx) const override final {
        assert(static_cast<std::size_t>(inputIdx) < NUM && "wrong inputIdx for Add operator.");
        return mInputs[inputIdx];
    }

    std::shared_ptr<Data> getOutput(IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return mOutput;
    }

    void setBackend(const std::string& name) {
        mImpl = Registrar<Add_Op<NUM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        for (std::size_t i = 0; i < NUM; ++i) {
            mInputs[i]->setBackend(name);
        }
    }

    void setDatatype(const DataType& datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        for (std::size_t i = 0; i < NUM; ++i) {
            mInputs[i]->setDatatype(datatype);
        }
    }

    IONb_t nbInputs() const override final { return NUM; }
    IONb_t nbDataInputs() const override final { return NUM; }
    IONb_t nbOutputs() const override final { return 1; }
};

template <std::size_t NUM>
inline std::shared_ptr<Node> Add(const char* name = nullptr) {
    return std::make_shared<Node>(std::make_shared<Add_Op<NUM>>(), name);
}
}

#endif /* __AIDGE_ADD_H__ */
