/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_CONV_H__
#define __AIDGE_CONV_H__

#include <array>
#include <cmath>
#include <numeric>
#include <vector>

#include "data/Tensor.hpp"
#include "graph/Node.hpp"
#include "operator/Operator.hpp"
#include "operator/Producer.hpp"
#include "utils/Parameter.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"

namespace Aidge {
enum class ConvParam { StrideDims, DilationDims, InChannels, OutChannels, KernelDims };

template <DimIdx_t DIM>
class Conv_Op : public Operator,
                public Registrable<std::string, std::unique_ptr<OperatorImpl>(const Conv_Op<DIM> &)>,
                public Parameterizable<ConvParam, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>, DimSize_t,
                                       DimSize_t, std::array<DimSize_t, DIM>> {
   public:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, 3> mInputs = {std::make_shared<Tensor>(), std::make_shared<Tensor>(),
                                                      std::make_shared<Tensor>()};
    std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

   public:
    static constexpr const char *Type = "Conv";

    Conv_Op() = delete;

    using Parameterizable_ = Parameterizable<ConvParam, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>,
                                             DimSize_t, DimSize_t, std::array<DimSize_t, DIM>>;
    template <ConvParam e>
    using param = typename Parameterizable_::template param<e>;

    constexpr Conv_Op(DimSize_t in_channels, DimSize_t out_channels, const std::array<DimSize_t, DIM> &kernel_dims)
        : Operator(Type),
          Parameterizable_(param<ConvParam::StrideDims>({}), param<ConvParam::DilationDims>({}),
                           param<ConvParam::InChannels>(in_channels), param<ConvParam::OutChannels>(out_channels),
                           param<ConvParam::KernelDims>(kernel_dims)),
          mOutput(std::make_shared<Tensor>()) {
        setDatatype(DataType::Float32);
        this->template get<ConvParam::StrideDims>().fill(1);
        this->template get<ConvParam::DilationDims>().fill(1);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    constexpr void setInput(IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    constexpr void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            std::array<DimSize_t, DIM + 2> outputDims = {};

            for (std::size_t dim = 0; dim < this->template get<ConvParam::KernelDims>().size() ; ++dim) {
                const DimSize_t kernelExtent = this->template get<ConvParam::DilationDims>()[dim] *
                                                       (this->template get<ConvParam::KernelDims>()[dim] - 1) +
                                               1;

                outputDims[dim+2] = static_cast<DimSize_t>(
                        floor(static_cast<float>(mInputs[0]->dims()[dim+2] - kernelExtent +
                                                 this->template get<ConvParam::StrideDims>()[dim]) /
                              static_cast<float>(this->template get<ConvParam::StrideDims>()[dim])));
            }

            outputDims[1] = this->template get<ConvParam::OutChannels>();
            outputDims[0] = mInputs[0]->dims()[0];
            mOutput->resize(outputDims);
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutput->empty()); }

    std::shared_ptr<Data> getInput(IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        return mInputs[inputIdx];
    }

    std::shared_ptr<Data> getOutput(IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return mOutput;
    }

    void setBackend(const std::string &name) {
        mImpl = Registrar<Conv_Op<DIM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[1]->setBackend(name);
        mInputs[2]->setBackend(name);
    }

    void setDatatype(const DataType &datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[0]->setDatatype(datatype);
        mInputs[1]->setDatatype(datatype);
        mInputs[2]->setDatatype(datatype);
    }

    IONb_t nbInputs() const override final { return 3; }
    IONb_t nbDataInputs() const override final { return 1; }
    IONb_t nbOutputs() const override final { return 1; }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Conv(DimSize_t in_channels, DimSize_t out_channels,
                                  const std::array<DimSize_t, DIM> &kernel_dims, const char *name = nullptr) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    auto conv = std::make_shared<Node>(std::make_shared<Conv_Op<static_cast<DimIdx_t>(DIM)>>(in_channels, out_channels, kernel_dims), name);
    // addProducer(conv, 1, append(append(kernel_dims, in_channels), out_channels), "w");
    addProducer(conv, 1, append(out_channels, append(in_channels, kernel_dims)), "w");
    addProducer(conv, 2, {out_channels}, "b");
    return conv;
}

template <DimSize_t DIM>
inline std::shared_ptr<Node> Conv(
    DimSize_t in_channels,
    DimSize_t out_channels,
    DimSize_t const (&kernel_dims)[DIM],
    const char *name = nullptr) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    return Conv(in_channels, out_channels, to_array(kernel_dims), name);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ConvParam>::data[] = {"StrideDims", "DilationDims", "InChannels", "OutChannels",
                                                          "KernelDims"};
}

#endif /* __AIDGE_CONV_H__ */
