/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_GENERIC_OPERATOR_H__
#define __AIDGE_GENERIC_OPERATOR_H__

#include <iostream>

#include "graph/Node.hpp"
#include "operator/Operator.hpp"
#include "utils/CParameter.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"

namespace Aidge {
class GenericOperator_Op
    : public Operator,
      public Registrable<std::string, std::unique_ptr<OperatorImpl>(std::shared_ptr<GenericOperator_Op>)> {
   private:
    CParameter mParams;
    IONb_t mNbDataIn;
    IONb_t mNbIn;
    IONb_t mNbOut;

   public:
    GenericOperator_Op(const char *type, IONb_t nbDataIn, IONb_t nbIn, IONb_t nbOut)
        : Operator(type), mNbDataIn(nbDataIn), mNbIn(nbIn), mNbOut(nbOut) {
        // ctor
    }

    /**
     * @brief Get the Parameter object identified by its name.
     * @tparam T expected parameter type.
     * @param key Parameter name.
     * @details assert if T is not the actual parameter type, if the parameter
     * does not exist or internal parameter position is invalid.
     * @todo Returning a T const& ? But dangerous => may get an address within
     * param buffer that will get invalid after the CParam death.
     * @note at() throws if the parameter does not exist, using find to test
     * for parameter existance
     * @return template<class T> The parameter.
     */
    template <class T>
    T getParameter(std::string const &key) const {
        return mParams.Get<T>(key);
    }

    ///\brief Add a parameter value, identified by its name
    ///\tparam T expected parameter type
    ///\param i_ParamName Parameter name
    ///\param i_Value Parameter value
    ///\todo Pass i_Value by ref if large or not trivial
    ///\bug If parameter already exists, its value is changed but written in the
    /// internal buffer in a new location (previous value is still in memory at
    /// its previous location)
    template <class T>
    void addParameter(std::string const &key, T const &value) {
        mParams.Add<T>(key, value);
    }

    std::string getParameterType(std::string const &key) { return mParams.getParamType(key); }

    std::vector<std::string> getParametersName() { return mParams.getParametersName(); }

    // Override Virtual Opertor methods
    void setInput(IOIndex_t /*inputIdx*/, std::shared_ptr<Data> /*data*/) override final {
        printf("Not available yet.\n");
    }

    void computeOutputDims() override final { printf("Not available yet"); }

    bool outputDimsForwarded() const override final {
        assert(false && "GenericOperator cannot forward dims");
        return false;
    }

    std::shared_ptr<Data> getInput(IOIndex_t /*inputIdx*/) const override final {
        printf("Not available yet.\n");
        return nullptr;
    }

    std::shared_ptr<Data> getOutput(IOIndex_t /*outputIdx*/) const override final {
        printf("Not available yet.\n");
        return nullptr;
    }
    ~GenericOperator_Op() = default;

    void setBackend(const std::string & /*name*/) { printf("Not available yet.\n"); }
    void setDatatype(const DataType & /*datatype*/) { printf("Not available yet.\n"); }
    void forward() override final { printf("Not available yet.\n"); }
    void backward() override final { printf("Not available yet.\n"); }

    IONb_t nbInputs() const override final { return mNbIn; };
    IONb_t nbDataInputs() const override final { return mNbDataIn; };
    IONb_t nbOutputs() const override final { return mNbOut; };
};

/**
 * @brief Fictive custom operator not associated with any implementation.
 * Allows to import unknown operators and simulate new ones.
 * @param type Type of the fictive operator.
 * @param nbDataIn Number of input data.
 * @param nbIn Number input data + number of learnt parameters.
 * @param nbOut Number of output data.
 * @param name (optional) name of the Operator.
 * @return std::shared_ptr<Node> Node associated with the Generic Operator.
 */
inline std::shared_ptr<Node> GenericOperator(const char *type, IONb_t nbDataIn, IONb_t nbIn, IONb_t nbOut,
                                             const char *name = nullptr) {
    return std::make_shared<Node>(std::make_shared<GenericOperator_Op>(type, nbDataIn, nbIn, nbOut), name);
}
}  // namespace Aidge

#endif /* __AIDGE_GENERIC_OPERATOR_H__ */
