/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_OPERATOR_H__
#define __AIDGE_OPERATOR_H__

#include <memory>
#include <string>
#include <vector>

#include "backend/OperatorImpl.hpp"
#include "data/Data.hpp"
#include "data/Tensor.hpp"
#include "utils/Types.h"

namespace Aidge {

class Operator : public std::enable_shared_from_this<Operator> {
protected:
    std::unique_ptr<OperatorImpl> mImpl; // implementation of the operator

private:
    std::string mType;

public:
    Operator() = delete;
    Operator(const char* type) : mType(type) {}
    virtual ~Operator();


public:

    virtual void setInput(IOIndex_t inputIdx, std::shared_ptr<Data> data) = 0;
    virtual void computeOutputDims() = 0;
    virtual bool outputDimsForwarded() const = 0;
    virtual std::shared_ptr<Data> getInput(IOIndex_t inputIdx) const = 0;
    virtual std::shared_ptr<Data> getOutput(IOIndex_t outputIdx) const = 0;

///////////////////////////////////////////////////////
//        IMPLEMENTATION
///////////////////////////////////////////////////////

    virtual void setBackend(const std::string& name) = 0;
    virtual void setDatatype(const DataType& datatype) = 0;

    /**
     * @brief Minimum amount of data from a specific input for one computation pass.
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t 
     */
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data from a specific input actually used in one computation pass.
     * 
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t 
     */
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data ready to be used on a specific output.
     * 
     * @param outputIdx Index of the output analysed.
     * @return NbElts_t 
     */
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const;

    virtual void forward();

    virtual void backward();

///////////////////////////////////////////////////////
//        INNER
///////////////////////////////////////////////////////

    std::string type() const {
        return mType;
    }

    virtual IONb_t nbInputs() const = 0;
    virtual IONb_t nbDataInputs() const = 0;
    virtual IONb_t nbOutputs() const = 0;
};
} // namespace Aidge

#endif /* __AIDGE_OPERATOR_H__ */
