/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_PRODUCER_H__
#define __AIDGE_PRODUCER_H__

#include <array>
#include <vector>

#include "utils/Types.h"
#include "data/Tensor.hpp"
#include "graph/Node.hpp"
#include "operator/Operator.hpp"
#include "utils/Parameter.hpp"
#include "utils/Registrar.hpp"

namespace Aidge {
enum class ProducerParam { Dims };

template <std::size_t DIM>
class Producer_Op
    : public Operator,
      public Registrable<std::string, std::unique_ptr<OperatorImpl>(
                                          const Producer_Op<DIM> &)>,
      public Parameterizable<ProducerParam, std::array<DimSize_t, DIM>> {
private:
    std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char* Type = "Producer";

    using Parameterizable_ = Parameterizable<ProducerParam, std::array<DimSize_t, DIM>>;
    template <ProducerParam e> using param = typename Parameterizable_::template param<e>;

    Producer_Op(const std::array<DimSize_t, DIM>& dims)
        : Operator(Type),
          Parameterizable_(
              param<ProducerParam::Dims>(dims)),
          mOutput(std::make_shared<Tensor>())
    {
        //ctor
        setDatatype(DataType::Float32);
        mOutput->resize(this->template get<ProducerParam::Dims>());
    }

    void setInput(IOIndex_t /*inputIdx*/, std::shared_ptr<Data> /*data*/) override final {
        assert(false && "Producer operator takes no input");
    }

    constexpr void computeOutputDims() override final {}

    constexpr bool outputDimsForwarded() const override final {return true;}

    std::shared_ptr<Data> getInput(IOIndex_t /*inputIdx*/) const override final {
        assert(false && "Producer operator takes no input");
        return nullptr;
    }

    std::shared_ptr<Data> getOutput(IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return mOutput;
    }

    void setOutput(std::shared_ptr<Data> outputData) {
        mOutput = std::static_pointer_cast<Aidge::Tensor>(outputData);
    }

    void setBackend(const std::string& name) {
        mImpl = Registrar<Producer_Op>::create(name)(*this);
        mOutput->setBackend(name);
    }
    void setDatatype(const DataType& datatype) {
        mOutput->setDatatype(datatype);
    }

    IONb_t nbInputs() const override final { return 0; };
    IONb_t nbDataInputs() const override final { return 0; };
    IONb_t nbOutputs() const override final { return 1; };

public:
  void forward() override final {
    printf("Basic Producer forward() function.\n");
  }
  void backward() override final {
    printf("Basic Producer backward() function.\n");
  }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Producer(const std::array<DimSize_t, DIM> &dims, const char *name = nullptr) {
  static_assert(DIM<=MaxDim,"Too many tensor dimensions required by Producer, not supported");
  return std::make_shared<Node>(std::make_shared<Producer_Op<DIM>>(dims), name);
}

template <std::size_t DIM>
inline std::shared_ptr<Node> Producer(DimSize_t const (&dims)[DIM], const char *name = nullptr) {
  return Producer(to_array(dims), name);
}

template <std::array<DimSize_t, 1>::size_type DIM>
void addProducer(std::shared_ptr<Node>& otherNode, const IOIndex_t inputIdx, const std::array<DimSize_t, DIM>& dims, const char* extension) {
    assert(inputIdx != gk_IODefaultIndex);
    static_assert(DIM<=MaxDim,"Too many tensor dimensions required by addProducer, not supported");
    const char* prodName = otherNode->name().empty() ? nullptr : (otherNode->name() + std::string("_") + std::string(extension)).c_str();
    auto prod = Producer(dims, prodName);
    prod->addChild(otherNode, 0, inputIdx);
    otherNode->getOperator()->setInput(inputIdx, std::static_pointer_cast<Producer_Op<DIM>>(prod->getOperator())->getOutput(0));
}

template <std::size_t DIM>
void addProducer(std::shared_ptr<Node>& otherNode, const IOIndex_t inputIdx, DimSize_t const (&dims)[DIM], const char* extension) {
    addProducer(otherNode, inputIdx, to_array(dims), extension);
}
} // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ProducerParam>::data[] = {"Dims"};
}

#endif /* __AIDGE_PRODUCER_H__ */