/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_RELU_H__
#define __AIDGE_RELU_H__

#include <vector>
#include <memory>

#include "utils/Parameter.hpp"
#include "utils/Registrar.hpp"
#include "operator/Operator.hpp"
#include "backend/OperatorImpl.hpp"
#include "data/Tensor.hpp"
#include "data/Data.hpp"
#include "graph/Node.hpp"
#include "utils/Types.h"

namespace Aidge {

enum class ReLUParam {
    Alpha
};

class ReLU_Op : public Operator,
    public Registrable<std::string, std::unique_ptr<OperatorImpl>(const ReLU_Op&)>,
    public Parameterizable<ReLUParam, float> {
public:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, 1> mInputs = {std::make_shared<Tensor>()};
    std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char* Type = "ReLU";

    ReLU_Op() = delete;

    using Parameterizable_ = Parameterizable<ReLUParam, float>;
    template <ReLUParam e> using param = typename Parameterizable_::template param<e>;

    ReLU_Op(float alpha)
            : Operator(Type),
            Parameterizable_(
                param<ReLUParam::Alpha>(alpha)),
            mOutput(std::make_shared<Tensor>())
    {
        setDatatype(DataType::Float32);
    }

    void setInput(IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx == 0 && "operator supports only 1 input");
        assert(strcmp(data->type(), Tensor::Type)==0 && "input data must be of Tensor type");
        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        if (!mInputs[0]->empty())
            mOutput->resize(mInputs[0]->dims());
    }

    bool outputDimsForwarded() const override final {
        return !(mOutput->empty());
    }

    std::shared_ptr<Data> getInput(IOIndex_t inputIdx) const override final {
        assert(inputIdx == 0 && "operator supports only 1 input");
        return mInputs[inputIdx];
    }

    std::shared_ptr<Data> getOutput(IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return mOutput;
    }

    void setBackend(const std::string& name) {
        mImpl = Registrar<ReLU_Op>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[0]->setBackend(name);
    }
    void setDatatype(const DataType& datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[0]->setDatatype(datatype);
    }

    IONb_t nbInputs() const override final { return 1; }
    IONb_t nbDataInputs() const override final { return 1; }
    IONb_t nbOutputs() const override final { return 1; }
};

inline std::shared_ptr<Node> ReLU(float alpha = 0.0f, const char* name = nullptr) {
    // FIXME: properly handle default w&b initialization in every cases
    return std::make_shared<Node>(std::make_shared<ReLU_Op>(alpha), name);
}

}

namespace {
template <>
const char* const EnumStrings<Aidge::ReLUParam>::data[]
    = {"Alpha"};
}

#endif /* __AIDGE_RELU_H__ */
