/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_SCHEDULER_H__
#define __AIDGE_SCHEDULER_H__

#include <chrono>
#include <memory>
#include <set>
#include <string>
#include <vector>

namespace Aidge {
class Node;
class GraphView;

class SequentialScheduler {
public:
    struct SchedulingElement {
        SchedulingElement(
            std::shared_ptr<Node> node_,
            std::chrono::time_point<std::chrono::high_resolution_clock> start_,
            std::chrono::time_point<std::chrono::high_resolution_clock> end_)
            : node(node_), start(start_), end(end_) {}

        std::shared_ptr<Node> node;
        std::chrono::time_point<std::chrono::high_resolution_clock> start;
        std::chrono::time_point<std::chrono::high_resolution_clock> end;
    };

    SequentialScheduler(std::shared_ptr<GraphView> graphView)
        : mGraphView(graphView)
    {
        // ctor
    };
    ~SequentialScheduler() = default;

    /**
     * @brief Run the provided Computational Graph with a batch of data
     */
    void forward(bool forwardDims = true, bool verbose = false);

    /**
     * @brief Save in a Markdown file the order of layers execution.
     * @param fileName Name of the generated file.
     */
    void saveSchedulingDiagram(const std::string& fileName) const;

private:
    /**
     * @brief Set of layers receiving an input from currently processing layers
     * 
     * @param producers Set of layers ready to run.
     * @return std::set<std::shared_ptr<Node>> 
     */
    std::set<std::shared_ptr<Node>> getConsumers(const std::set<std::shared_ptr<Node>>& producers) const;

    std::shared_ptr<GraphView> mGraphView;
    std::vector<SchedulingElement> mScheduling;
};
} // namespace Aidge

#endif /* __AIDGE_SCHEDULER_H__ */