/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_CPARAMETER_H__
#define __AIDGE_CPARAMETER_H__

#include <assert.h>
#include <map>
#include <vector>
#include <numeric>

namespace Aidge {

///\todo store also a fix-sized code that indicates the type
///\todo managing complex types or excluding non-trivial, non-aggregate types
class CParameter
{
public:
    // not copyable, not movable
    CParameter(CParameter const &) = delete;
    CParameter(CParameter &&) = delete;
    CParameter &operator=(CParameter const &) = delete;
    CParameter &operator=(CParameter &&) = delete;
    CParameter() : m_Params({}){};

    /**
     * \brief Returning a parameter identified by its name
     * \tparam T expected parameter type
     * \param i_ParamName Parameter name
     * \details assert if T is not the actual parameter type, if the parameter does not
     *  exist or interna parameter position is invalid.
     * \todo Returning a T const& ? But dangerous => the client may get an address within
     *  param buffer that will get invalid after the CParam death.
     * \note at() throws if the parameter does not exist, using find to test for parameter existance
     */
    template<class T> T Get(std::string const &i_ParamName) const
    {
        assert(m_Params.find(i_ParamName) != m_Params.end());
        assert(m_Types.find(i_ParamName) != m_Types.end());
        assert(m_Params.at(i_ParamName) <= m_OffSet);
        assert(typeid(T).name() == m_Types.at(i_ParamName));
        return *reinterpret_cast<T *>(m_BeginBuffer + m_Params.at(i_ParamName));
    }

    ///\brief Add a parameter value, identified by its name
    ///\tparam T expected parameter type
    ///\param i_ParamName Parameter name
    ///\param i_Value Parameter value
    ///\todo Pass i_Value by ref if large or not trivial
    ///\bug If parameter already exists, its value is changed but written in the
    /// internal buffer in a new location (previous value is still in memory at its previous location)
    template<class T> void Add(std::string const &i_ParamName, T const &i_Value)
    {
        m_Buffer.resize(m_Buffer.size() + (sizeof(T) / sizeof(uint8_t)));
        m_BeginBuffer = m_Buffer.data(); // Update buffer ptr in case of memory reordering
        *reinterpret_cast<T *>(m_BeginBuffer + m_OffSet)
            = i_Value; // Black-magic used to add anytype into the vector
        m_Params[i_ParamName] = m_OffSet; // Copy pointer offset
        m_OffSet += sizeof(T); // Increment offset
        m_Types[i_ParamName] = typeid(i_Value).name();
    }
    
    std::string getParamType(std::string const &i_ParamName){
        return m_Types[i_ParamName];
    }

    std::vector<std::string> getParametersName(){
        std::vector<std::string> parametersName;
        for(auto const& it: m_Params)
            parametersName.push_back(it.first);
        return parametersName;
    }


    ~CParameter() = default;

private:
    // Note for Cyril: of course storing offset and not address! Good idea
    std::map<std::string, std::size_t> m_Params; // { Param name : offset }

    ///\brief Map to check type error
    /* Note : i tried this : `std::map<std::string, std::type_info const *> m_Types;`
    but looks like the type_ingo object was destroyed.
    I am not a hugde fan of storing a string and making string comparison.
    Maybe we can use a custom enum type (or is there a standard solution ?)  
    */
    std::map<std::string, std::string> m_Types;

    ///\brief All parameters values concatenated in raw binary form.
    std::vector<uint8_t> m_Buffer = {};

    ///\brief Starting address of the buffer
    uint8_t *m_BeginBuffer = m_Buffer.data();

    ///\brief Offset, in number of uint8_t, of the next parameter to write
    std::size_t m_OffSet = 0;
};

}


#endif /* __AIDGE_CPARAMETER_H__ */
