/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <string>

#include "utils/Parameter.hpp"
#include "backend/OperatorImpl.hpp"
#include "operator/Conv.hpp"
#include "operator/Operator.hpp"
#include "utils/Types.h"

namespace py = pybind11;
namespace Aidge {

template <DimIdx_t DIM> void declare_ConvOp(py::module &m) {
  py::class_<Conv_Op<DIM>, std::shared_ptr<Conv_Op<DIM>>, Operator>(
    m, ("ConvOp" + std::to_string(DIM) + "D").c_str(),
    py::multiple_inheritance())
  .def(py::init<DimSize_t, DimSize_t, const std::array<DimSize_t, DIM> &>(),
        py::arg("in_channels"), py::arg("out_channels"),
        py::arg("kernel_dims"));
  
  m.def(("Conv" + std::to_string(DIM) + "D").c_str(), [](DimSize_t in_channels, DimSize_t out_channels, std::vector<DimSize_t>& kernel_dims, const char* name) {
        // Lambda function wrapper because PyBind fails to convert const array.
        // So we use a vector that we convert in this function to a const DimeSize_t [DIM] array. 
        if (kernel_dims.size() != DIM) {
            throw std::runtime_error("kernel_dims size [" + std::to_string(kernel_dims.size()) + "] does not match DIM [" + std::to_string(DIM) +"]");
        }
        DimSize_t tmp_kernel_dims_array[DIM];
        for (size_t i = 0; i < DIM; ++i) {
            tmp_kernel_dims_array[i] = kernel_dims[i];
        }
        const DimSize_t (&kernel_dims_array)[DIM] = tmp_kernel_dims_array;
        return Conv<DIM>(in_channels, out_channels, kernel_dims_array, name);
    }, py::arg("in_channels"), py::arg("out_channels"),
        py::arg("kernel_dims"), py::arg("name") = nullptr);
  
}


void init_Conv(py::module &m) {
  declare_ConvOp<1>(m);
  declare_ConvOp<2>(m);
  declare_ConvOp<3>(m);
 
  // FIXME:
  // m.def("Conv1D", static_cast<NodeAPI(*)(const char*, int, int, int const
  // (&)[1])>(&Conv));
}
} // namespace Aidge
