/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "graphmatching/GRegex.hpp"
#include "graph/GraphView.hpp"

using namespace Aidge; 

GRegex::GRegex(const std::map<std::string,NodeRegex*>& nodesRegex,std::vector<std::string>& seqRegexps ):mStmFab(nodesRegex){

 
    //setup all the STM
    for (const std::string& sequRegex : seqRegexps) {
        mStmInit.push_back(mStmFab.makeNewStm(sequRegex));
    }
    
}

bool GRegex::walk_validation_all_stm_are_valid(const std::vector<std::vector<SeqStm*>> all_stm){
    //test if all stm type are in a valid state
    std::vector<int> number_of_valid;
    number_of_valid.resize(all_stm.size());

    for (std::size_t i = 0; i < all_stm.size(); ++i) {
        number_of_valid[i] = 0;
        for (auto it = all_stm[i].begin(); it != all_stm[i].end(); ++it) {
            SeqStm* stm = *it;
            if (stm->isValid()){ 
                number_of_valid[i] +=1;
            }
        }   
    }

    for (std::size_t i = 0; i < number_of_valid.size(); ++i) {
        if (number_of_valid[i] == 0) {
            //std::cout << "NO MATCH at least one stm are not valid" << std::endl;
            return false;
        }
        if (number_of_valid[i] > 1) {
            //std::cout << "NO MATCH multiple brach match of stm (// quantification)" << std::endl;
            return false;
        }
    }
    return true;
}

bool GRegex::walk_validation_all_node_read_validate_by_one_stm(const std::vector<std::vector<SeqStm*>> all_stm){
    std::set<NodeTmp> all_stm_node_tested;
    std::set<NodeTmp> all_stm_node_validated;

    for (std::size_t i = 0; i < all_stm.size(); ++i) {
        //std::cout << "all stm index " << i <<  " on dimension 1 of size " << all_stm.size() <<std::endl;
        for (std::size_t j = 0; j < all_stm[i].size(); ++j) {
            //std::cout << "all stm index " << j <<  " on dimension 2 of size " << all_stm[i].size() <<std::endl;

            std::set<NodeTmp> stm_node_tested = all_stm[i][j]->getAllNodeTested();
            std::set<NodeTmp> stm_node_validated = all_stm[i][j]->getAllNodeValidated();

            all_stm_node_tested.insert(stm_node_tested.begin(), stm_node_tested.end()); 
            all_stm_node_validated.insert(stm_node_validated.begin(), stm_node_validated.end()); 
        }   
    }
    

    std::set<NodeTmp> test_but_not_valid;
    for (const auto& x : all_stm_node_tested) {
        if (all_stm_node_validated.find(x) == all_stm_node_validated.end()) {
            test_but_not_valid.insert(x);
        }
    }


    if (!test_but_not_valid.empty()) {
        std::cout << "NO MATCH. The node(s) ";
        for (const auto& x : test_but_not_valid) {
            std::cout << x.get() << ", ";
        }
        std::cout << " have been tested but not validated." << std::endl;
        return false;
    }
    return true;

}

bool GRegex::walk_validation_common_nodes_same_tag_for_all_stm(const std::vector<std::vector<SeqStm*>> all_stm){
    std::map<NodeTmp, std::pair<std::string,int>> node_to_common_tag;
    for (std::size_t i = 0; i < all_stm.size(); ++i) {
        for (auto it = all_stm[i].begin(); it != all_stm[i].end(); ++it) {
            SeqStm* stm = *it;
            
            if (!stm->isValid()){ 
                continue;
            }
            
            for (const auto& pair : stm->getAllCommonNode()) {
                const NodeTmp node = pair.first;
                const std::string common_tag = pair.second;

                if (node_to_common_tag.find(node) != node_to_common_tag.end()) {
                    std::string tag = node_to_common_tag[node].first;
                    int& occurence = node_to_common_tag[node].second;
                    if (tag!=common_tag){
                        std::cout << "NO MATCH. The node " << node << " have two different tags "<< tag << " and " << common_tag  << std::endl;
                        return false;
                    } else {
                        occurence += 1;
                    }
                } else {
                    node_to_common_tag.insert(std::make_pair(node, std::make_pair(common_tag, 1)));
                }
            }
        }   
    }
    /*std::cout << "Node to common tag ";
    for (const auto& x : node_to_common_tag) {
        std::cout << "(" << x.first << ", " << "[" << x.second.first << ", " << x.second.second << "]" << ") ; ";
    }
    std::cout << std::endl;*/


    for (const auto& pair : node_to_common_tag) {
        const std::pair<std::string, int> tag_occurence_pair = pair.second;
        if (tag_occurence_pair.second < 1){
            //std::cout << "NO MATCH. The common tag " << tag_occurence_pair.first  << " did not match " << std::endl;
            return false;
        }
    }
    
    return true;
}

std::set<NodeTmp> GRegex::get_all_validate_nodes(const std::vector<std::vector<SeqStm*>> all_stm){
    std::set<NodeTmp> all_stm_node_validated;

    for (std::size_t i = 0; i < all_stm.size(); ++i) {
        for (std::size_t j = 0; j < all_stm[i].size(); ++j) {
            std::set<NodeTmp> stm_node_validated = all_stm[i][j]->getAllNodeValidated();
            all_stm_node_validated.insert(stm_node_validated.begin(), stm_node_validated.end()); 
        }   
    }
    return all_stm_node_validated;
}


std::set<NodeTmp> GRegex::matchFromStartNodes(const std::vector<NodeTmp> startNodes,const std::shared_ptr<GraphView> graphToMatch){
    std::set<NodeTmp> empty_set_return;
    //ASSERT
    if(startNodes.size() != mStmInit.size()){
        throw std::runtime_error ("bad GRegex start nodes");
    }

        //init the walk
        std::vector<std::vector<SeqStm*>> allStm;
        std::vector<std::pair<NodeTmp,SeqStm*>> currentWalk;

        for (SeqStm* seqStmPtr : mStmInit) {
            SeqStm* newStm = mStmFab.duplicateStm(seqStmPtr);
            std::size_t idxStart = newStm->getStmIdx();
            currentWalk.push_back(std::make_pair(startNodes[idxStart],newStm));
            allStm.push_back(std::vector<SeqStm*>());
        }

        //walk
        while (currentWalk.size()!=0)
        {
            std::vector<std::pair<NodeTmp,SeqStm*>> newWalk;
            for (const auto& pair : currentWalk) {
                const NodeTmp node = pair.first;
                SeqStm* stmPtr = pair.second;

                std::pair<int,std::string> test = stmPtr->testNode(node);
                int res = test.first;
                std::string commonTag = test.second;

                std::set<NodeTmp> next_nodes = graphToMatch->getChildren(node);

                /*std::cout << "Next nodes : " ;
                for (const auto& x : next_nodes) {
                    std::cout << x->name() << ", ";
                }
                std::cout << std::endl;*/
                
                // Test Match
                if (commonTag == "" && next_nodes.size() > 1) {
                    std::cout << "NO MATCH. The node " << node.get() << " is not common and has more than one child" << std::endl;
                    return empty_set_return;
                }

                // If there is no more nodes --> Archive the branch
                if (res == -1 || next_nodes.empty()) {
                    int indexToInsert = stmPtr->getStmIdx();
                    allStm[indexToInsert].push_back(stmPtr);
                    //std::cout << "No more nodes --> STM archived : " << indexToInsert << std::endl;
                    continue; // TODEV : replace this with 'else' that encapsulate the rest of the function ? 
                }

                bool first = true;

                // Use an iterator to read through the next_nodes
                std::set<NodeTmp>::iterator it;
                for (it = next_nodes.begin(); it != next_nodes.end(); ++it) {
                    // Access the current element using the iterator
                    std::shared_ptr<Aidge::Node> next_node = *it;
                    if (first){
                        newWalk.push_back(std::make_pair(next_node, stmPtr));
                        first = false;
                    } else {
                        SeqStm* new_stmPtr = mStmFab.duplicateStm(stmPtr);
                        newWalk.push_back(std::make_pair(next_node, new_stmPtr));
                    }
                }
            }
            currentWalk = newWalk;
        }
    
    //std::cout << "Walk finished" << std::endl;    

    if (!walk_validation_all_stm_are_valid(allStm)){
        return empty_set_return;
    }
    //std::cout << "walk_validation_all_stm_are_valid finished" << std::endl;
    

    if (!walk_validation_all_node_read_validate_by_one_stm(allStm)){
        return empty_set_return;
    }
    //std::cout << "walk_validation_all_node_read_validate_by_one_stm finished" << std::endl;
    

    if (!walk_validation_common_nodes_same_tag_for_all_stm(allStm)){
        return empty_set_return;
    }
    //std::cout << "walk_validation_common_nodes_same_tag_for_all_stm finished" << std::endl;

    //std::cout << "MATCH" << std::endl;
    
    return get_all_validate_nodes(allStm);
        
}



Match GRegex::match(const std::shared_ptr<GraphView> graphToMatch){

    //std::vector<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>>  matches;
    //std::set<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>>  matches;
    Match  matches;
    std::size_t nbStartNodes = mStmInit.size();
    std::set<NodeTmp> allNodes = graphToMatch->getNodes();
    std::size_t nbAllNodes = allNodes.size();

    std::vector<std::size_t> indices(nbStartNodes, 0);

       while (true) {
        // Generate all permutations of the current combination
        do {
            std::vector<NodeTmp> startNodes;
            //std::cout <<"start nodes :";
            for (std::size_t i = 0; i < nbStartNodes; ++i) {
                auto it = std::begin(allNodes);
                std::advance(it, indices[i]);
                //std::cout << (*it).get() << " ";
                startNodes.push_back(*it);
            }
            //std::cout <<"\n";

           std::set<NodeTmp> match =  matchFromStartNodes(startNodes, graphToMatch);
           //std::cout << "match size : " << match.size() << " ";
           if(match.size() != 0){
                //matches.push_back(std::make_pair(startNodes,match));
                //matches.insert(std::make_pair(startNodes,match));
                matches.insert(startNodes,match);
           }
            
        } while (std::next_permutation(indices.begin(), indices.end()));
        
        // Generate the next combination with replacement
        std::size_t i = nbStartNodes - 1;
        while (true) {
            if (indices[i] < nbAllNodes - 1) {
                ++indices[i];
                break;
            }
            if (i == 0) {
                return matches;
            }
            --i;
        }
        std::fill(indices.begin() + i + 1, indices.end(), indices[i]);
    }

    return matches;
}